@@ -1,6 +1,6 @@
 """Scan primitive."""
+from __future__ import annotations
 
-from __future__ import division, absolute_import
 
 __copyright__ = """
 Copyright 2011-2012 Andreas Kloeckner
@@ -12,7 +12,7 @@ Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
-    http://www.apache.org/licenses/LICENSE-2.0
+    https://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
@@ -20,30 +20,39 @@ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 
-Derived from code within the Thrust project, https://github.com/thrust/thrust/
+Derived from code within the Thrust project, https://github.com/NVIDIA/thrust
 """
 
-import six
-from six.moves import range, zip
+import logging
+from abc import ABC, abstractmethod
+from dataclasses import dataclass
+from typing import TYPE_CHECKING, Any, cast
 
 import numpy as np
 
-import pyopencl as cl
-import pyopencl.array  # noqa
-from pyopencl.tools import (dtype_to_ctype, bitlog2,
-        KernelTemplateBase, _process_code_for_macro,
-        get_arg_list_scalar_arg_dtypes,
-        context_dependent_memoize,
-        _NumpyTypesKeyBuilder,
-        get_arg_offset_adjuster_code)
-
-import pyopencl._mymako as mako
-from pyopencl._cluda import CLUDA_PREAMBLE
-
 from pytools.persistent_dict import WriteOncePersistentDict
 
+import pyopencl as cl
+import pyopencl._mymako as mako
+import pyopencl.array as cl_array
+from pyopencl._cluda import CLUDA_PREAMBLE
+from pyopencl.tools import (
+    DtypedArgument,
+    KernelTemplateBase,
+    _NumpyTypesKeyBuilder,
+    _process_code_for_macro,
+    bitlog2,
+    context_dependent_memoize,
+    dtype_to_ctype,
+    get_arg_list_scalar_arg_dtypes,
+    get_arg_offset_adjuster_code,
+)
+
+
+if TYPE_CHECKING:
+    from collections.abc import Sequence
+
 
-import logging
 logger = logging.getLogger(__name__)
 
 
@@ -746,7 +755,7 @@ void ${name_prefix}_final_update(
 
 # {{{ helpers
 
-def _round_down_to_power_of_2(val):
+def _round_down_to_power_of_2(val: int) -> int:
     result = 2**bitlog2(val)
     if result > val:
         result >>= 1
@@ -786,7 +795,7 @@ _IGNORED_WORDS = set("""
 
         typedef for endfor if void while endwhile endfor endif else const printf
         None return bool n char true false ifdef pycl_printf str range assert
-        np iinfo max itemsize __packed__ struct restrict
+        np iinfo max itemsize __packed__ struct restrict ptrdiff_t
 
         set iteritems len setdefault
 
@@ -840,10 +849,11 @@ _IGNORED_WORDS = set("""
         """.split())
 
 
-def _make_template(s):
+def _make_template(s: str):
+    import re
     leftovers = set()
 
-    def replace_id(match):
+    def replace_id(match: re.Match) -> str:
         # avoid name clashes with user code by adding 'psc_' prefix to
         # identifiers.
 
@@ -851,43 +861,29 @@ def _make_template(s):
         if word in _IGNORED_WORDS:
             return word
         elif word in _PREFIX_WORDS:
-            return "psc_"+word
+            return f"psc_{word}"
         else:
             leftovers.add(word)
             return word
 
-    import re
     s = re.sub(r"\b([a-zA-Z0-9_]+)\b", replace_id, s)
-
     if leftovers:
         from warnings import warn
-        warn("leftover words in identifier prefixing: " + " ".join(leftovers))
+        warn("Leftover words in identifier prefixing: " + " ".join(leftovers),
+             stacklevel=3)
 
     return mako.template.Template(s, strict_undefined=True)
 
 
-from pytools import Record, RecordWithoutPickling
+@dataclass(frozen=True)
+class _GeneratedScanKernelInfo:
+    scan_src: str
+    kernel_name: str
+    scalar_arg_dtypes: list[np.dtype | None]
+    wg_size: int
+    k_group_size: int
 
-
-class _GeneratedScanKernelInfo(Record):
-
-    __slots__ = [
-            "scan_src",
-            "kernel_name",
-            "scalar_arg_dtypes",
-            "wg_size",
-            "k_group_size"]
-
-    def __init__(self, scan_src, kernel_name, scalar_arg_dtypes, wg_size,
-            k_group_size):
-        Record.__init__(self,
-                scan_src=scan_src,
-                kernel_name=kernel_name,
-                scalar_arg_dtypes=scalar_arg_dtypes,
-                wg_size=wg_size,
-                k_group_size=k_group_size)
-
-    def build(self, context, options):
+    def build(self, context: cl.Context, options: Any) -> _BuiltScanKernelInfo:
         program = cl.Program(context, self.scan_src).build(options)
         kernel = getattr(program, self.kernel_name)
         kernel.set_scalar_arg_dtypes(self.scalar_arg_dtypes)
@@ -897,40 +893,33 @@ class _GeneratedScanKernelInfo(Record):
                 k_group_size=self.k_group_size)
 
 
-class _BuiltScanKernelInfo(RecordWithoutPickling):
-
-    __slots__ = ["kernel", "wg_size", "k_group_size"]
-
-    def __init__(self, kernel, wg_size, k_group_size):
-        RecordWithoutPickling.__init__(self,
-                kernel=kernel,
-                wg_size=wg_size,
-                k_group_size=k_group_size)
+@dataclass(frozen=True)
+class _BuiltScanKernelInfo:
+    kernel: cl.Kernel
+    wg_size: int
+    k_group_size: int
 
 
-class _GeneratedFinalUpdateKernelInfo(Record):
+@dataclass(frozen=True)
+class _GeneratedFinalUpdateKernelInfo:
+    source: str
+    kernel_name: str
+    scalar_arg_dtypes: Sequence[np.dtype | None]
+    update_wg_size: int
 
-    def __init__(self, source, kernel_name, scalar_arg_dtypes, update_wg_size):
-        Record.__init__(self,
-                source=source,
-                kernel_name=kernel_name,
-                scalar_arg_dtypes=scalar_arg_dtypes,
-                update_wg_size=update_wg_size)
-
-    def build(self, context, options):
+    def build(self,
+              context: cl.Context,
+              options: Any) -> _BuiltFinalUpdateKernelInfo:
         program = cl.Program(context, self.source).build(options)
         kernel = getattr(program, self.kernel_name)
         kernel.set_scalar_arg_dtypes(self.scalar_arg_dtypes)
         return _BuiltFinalUpdateKernelInfo(kernel, self.update_wg_size)
 
 
-class _BuiltFinalUpdateKernelInfo(RecordWithoutPickling):
-    __slots__ = ["kernel", "update_wg_size"]
-
-    def __init__(self, kernel, update_wg_size):
-        RecordWithoutPickling.__init__(self,
-                kernel=kernel,
-                update_wg_size=update_wg_size)
+@dataclass(frozen=True)
+class _BuiltFinalUpdateKernelInfo:
+    kernel: cl.Kernel
+    update_wg_size: int
 
 # }}}
 
@@ -939,14 +928,25 @@ class ScanPerformanceWarning(UserWarning):
     pass
 
 
-class _GenericScanKernelBase(object):
+class GenericScanKernelBase(ABC):
     # {{{ constructor, argument processing
 
-    def __init__(self, ctx, dtype,
-            arguments, input_expr, scan_expr, neutral, output_statement,
-            is_segment_start_expr=None, input_fetch_exprs=[],
-            index_dtype=np.int32,
-            name_prefix="scan", options=[], preamble="", devices=None):
+    def __init__(
+            self,
+            ctx: cl.Context,
+            dtype: Any,
+            arguments: str | list[DtypedArgument],
+            input_expr: str,
+            scan_expr: str,
+            neutral: str | None,
+            output_statement: str,
+            is_segment_start_expr: str | None = None,
+            input_fetch_exprs: list[tuple[str, str, int]] | None = None,
+            index_dtype: Any = None,
+            name_prefix: str = "scan",
+            options: Any = None,
+            preamble: str = "",
+            devices: Sequence[cl.Device] | None = None) -> None:
         """
         :arg ctx: a :class:`pyopencl.Context` within which the code
             for this scan kernel will be generated.
@@ -958,54 +958,54 @@ class _GenericScanKernelBase(object):
             specified. All types used here must be known to PyOpenCL.
             (see :func:`pyopencl.tools.get_or_register_dtype`).
         :arg scan_expr: The associative, binary operation carrying out the scan,
-            represented as a C string. Its two arguments are available as `a`
-            and `b` when it is evaluated. `b` is guaranteed to be the
-            'element being updated', and `a` is the increment. Thus,
+            represented as a C string. Its two arguments are available as ``a``
+            and ``b`` when it is evaluated. ``b`` is guaranteed to be the
+            'element being updated', and ``a`` is the increment. Thus,
             if some data is supposed to just propagate along without being
-            modified by the scan, it should live in `b`.
+            modified by the scan, it should live in ``b``.
 
             This expression may call functions given in the *preamble*.
 
-            Another value available to this expression is `across_seg_boundary`,
+            Another value available to this expression is ``across_seg_boundary``,
             a C `bool` indicating whether this scan update is crossing a
-            segment boundary, as defined by `is_segment_start_expr`.
+            segment boundary, as defined by ``is_segment_start_expr``.
             The scan routine does not implement segmentation
-            semantics on its own. It relies on `scan_expr` to do this.
-            This value is available (but always `false`) even for a
+            semantics on its own. It relies on ``scan_expr`` to do this.
+            This value is available (but always ``false``) even for a
             non-segmented scan.
 
             .. note::
 
                 In early pre-releases of the segmented scan,
                 segmentation semantics were implemented *without*
-                relying on `scan_expr`.
+                relying on ``scan_expr``.
 
         :arg input_expr: A C expression, encoded as a string, resulting
             in the values to which the scan is applied. This may be used
             to apply a mapping to values stored in *arguments* before being
             scanned. The result of this expression must match *dtype*.
-            The index intended to be mapped is available as `i` in this
+            The index intended to be mapped is available as ``i`` in this
             expression. This expression may also use the variables defined
             by *input_fetch_expr*.
 
             This expression may also call functions given in the *preamble*.
         :arg output_statement: a C statement that writes
-            the output of the scan. It has access to the scan result as `item`,
-            the preceding scan result item as `prev_item`, and the current index
-            as `i`. `prev_item` in a segmented scan will be the neutral element
+            the output of the scan. It has access to the scan result as ``item``,
+            the preceding scan result item as ``prev_item``, and the current index
+            as ``i``. ``prev_item`` in a segmented scan will be the neutral element
             at a segment boundary, not the immediately preceding item.
 
             Using *prev_item* in output statement has a small run-time cost.
-            `prev_item` enables the construction of an exclusive scan.
+            ``prev_item`` enables the construction of an exclusive scan.
 
             For non-segmented scans, *output_statement* may also reference
-            `last_item`, which evaluates to the scan result of the last
+            ``last_item``, which evaluates to the scan result of the last
             array entry.
         :arg is_segment_start_expr: A C expression, encoded as a string,
-            resulting in a C `bool` value that determines whether a new
+            resulting in a C ``bool`` value that determines whether a new
             scan segments starts at index *i*.  If given, makes the scan a
-            segmented scan. Has access to the current index `i`, the result
-            of *input_expr* as `a`, and in addition may use *arguments* and
+            segmented scan. Has access to the current index ``i``, the result
+            of *input_expr* as ``a``, and in addition may use *arguments* and
             *input_fetch_expr* variables just like *input_expr*.
 
             If it returns true, then previous sums will not spill over into the
@@ -1016,8 +1016,8 @@ class _GenericScanKernelBase(object):
 
                 ARG_NAME_TYPE NAME = ARG_NAME[i+OFFSET];
 
-            `OFFSET` is allowed to be 0 or -1, and `ARG_NAME_TYPE` is the type
-            of `ARG_NAME`.
+            ``OFFSET`` is allowed to be 0 or -1, and ``ARG_NAME_TYPE`` is the type
+            of ``ARG_NAME``.
         :arg preamble: |preamble|
 
         The first array in the argument list determines the size of the index
@@ -1029,7 +1029,14 @@ class _GenericScanKernelBase(object):
         being processed in the scan.
         """
 
-        self.context = ctx
+        if index_dtype is None:
+            index_dtype = np.dtype(np.int32)
+
+        if input_fetch_exprs is None:
+            input_fetch_exprs = []
+
+        self.context: cl.Context = ctx
+        self.dtype: np.dtype[Any]
         dtype = self.dtype = np.dtype(dtype)
 
         if neutral is None:
@@ -1042,35 +1049,35 @@ class _GenericScanKernelBase(object):
         if dtype.itemsize % 4 != 0:
             raise TypeError("scan value type must have size divisible by 4 bytes")
 
-        self.index_dtype = np.dtype(index_dtype)
+        self.index_dtype: np.dtype[np.integer] = np.dtype(index_dtype)
         if np.iinfo(self.index_dtype).min >= 0:
             raise TypeError("index_dtype must be signed")
 
         if devices is None:
             devices = ctx.devices
-        self.devices = devices
+        self.devices: Sequence[cl.Device] = devices
         self.options = options
 
         from pyopencl.tools import parse_arg_list
-        self.parsed_args = parse_arg_list(arguments)
+        self.parsed_args: Sequence[DtypedArgument] = parse_arg_list(arguments)
         from pyopencl.tools import VectorArg
-        self.first_array_idx = [
+        self.first_array_idx: int = next(
                 i for i, arg in enumerate(self.parsed_args)
-                if isinstance(arg, VectorArg)][0]
+                if isinstance(arg, VectorArg))
 
-        self.input_expr = input_expr
+        self.input_expr: str = input_expr
 
-        self.is_segment_start_expr = is_segment_start_expr
-        self.is_segmented = is_segment_start_expr is not None
-        if self.is_segmented:
+        self.is_segment_start_expr: str | None = is_segment_start_expr
+        self.is_segmented: bool = is_segment_start_expr is not None
+        if is_segment_start_expr is not None:
             is_segment_start_expr = _process_code_for_macro(is_segment_start_expr)
 
-        self.output_statement = output_statement
+        self.output_statement: str = output_statement
 
-        for name, arg_name, ife_offset in input_fetch_exprs:
+        for _name, _arg_name, ife_offset in input_fetch_exprs:
             if ife_offset not in [0, -1]:
                 raise RuntimeError("input_fetch_expr offsets must either be 0 or -1")
-        self.input_fetch_exprs = input_fetch_exprs
+        self.input_fetch_exprs: Sequence[tuple[str, str, int]] = input_fetch_exprs
 
         arg_dtypes = {}
         arg_ctypes = {}
@@ -1078,30 +1085,28 @@ class _GenericScanKernelBase(object):
             arg_dtypes[arg.name] = arg.dtype
             arg_ctypes[arg.name] = dtype_to_ctype(arg.dtype)
 
-        self.options = options
-        self.name_prefix = name_prefix
+        self.name_prefix: str = name_prefix
 
         # {{{ set up shared code dict
 
-        from pytools import all
         from pyopencl.characterize import has_double_support
 
-        self.code_variables = dict(
-            np=np,
-            dtype_to_ctype=dtype_to_ctype,
-            preamble=preamble,
-            name_prefix=name_prefix,
-            index_dtype=self.index_dtype,
-            scan_dtype=dtype,
-            is_segmented=self.is_segmented,
-            arg_dtypes=arg_dtypes,
-            arg_ctypes=arg_ctypes,
-            scan_expr=_process_code_for_macro(scan_expr),
-            neutral=_process_code_for_macro(neutral),
-            is_gpu=bool(self.devices[0].type & cl.device_type.GPU),
-            double_support=all(
+        self.code_variables = {
+            "np": np,
+            "dtype_to_ctype": dtype_to_ctype,
+            "preamble": preamble,
+            "name_prefix": name_prefix,
+            "index_dtype": self.index_dtype,
+            "scan_dtype": dtype,
+            "is_segmented": self.is_segmented,
+            "arg_dtypes": arg_dtypes,
+            "arg_ctypes": arg_ctypes,
+            "scan_expr": _process_code_for_macro(scan_expr),
+            "neutral": _process_code_for_macro(neutral),
+            "is_gpu": bool(self.devices[0].type & cl.device_type.GPU),
+            "double_support": all(
                 has_double_support(dev) for dev in devices),
-            )
+            }
 
         index_typename = dtype_to_ctype(self.index_dtype)
         scan_typename = dtype_to_ctype(dtype)
@@ -1128,24 +1133,31 @@ class _GenericScanKernelBase(object):
 
         # }}}
 
-        self.use_lookbehind_update = "prev_item" in self.output_statement
-        self.store_segment_start_flags = (
+        self.use_lookbehind_update: bool = "prev_item" in self.output_statement
+        self.store_segment_start_flags: bool = (
                 self.is_segmented and self.use_lookbehind_update)
 
         self.finish_setup()
 
     # }}}
 
-    def finish_setup(self):
-        raise NotImplementedError
+    @abstractmethod
+    def finish_setup(self) -> None:
+        pass
 
 
-generic_scan_kernel_cache = WriteOncePersistentDict(
-        "pyopencl-generated-scan-kernel-cache-v1",
-        key_builder=_NumpyTypesKeyBuilder())
+if not cl._PYOPENCL_NO_CACHE:
+    generic_scan_kernel_cache: WriteOncePersistentDict[Any,
+                    tuple[_GeneratedScanKernelInfo, _GeneratedScanKernelInfo,
+                    _GeneratedFinalUpdateKernelInfo]] = \
+        WriteOncePersistentDict(
+            "pyopencl-generated-scan-kernel-cache-v1",
+            key_builder=_NumpyTypesKeyBuilder(),
+            in_mem_cache_size=0,
+            safe_sync=False)
 
 
-class GenericScanKernel(_GenericScanKernelBase):
+class GenericScanKernel(GenericScanKernelBase):
     """Generates and executes code that performs prefix sums ("scans") on
     arbitrary types, with many possible tweaks.
 
@@ -1162,40 +1174,43 @@ class GenericScanKernel(_GenericScanKernelBase):
         a = cl.array.arange(queue, 10000, dtype=np.int32)
         knl(a, queue=queue)
 
+    .. automethod:: __init__
+    .. automethod:: __call__
     """
 
-    def finish_setup(self):
+    def finish_setup(self) -> None:
         # Before generating the kernel, see if it's cached.
         from pyopencl.cache import get_device_cache_id
         devices_key = tuple(get_device_cache_id(device)
                 for device in self.devices)
 
         cache_key = (self.kernel_key, devices_key)
-
         from_cache = False
 
-        try:
-            result = generic_scan_kernel_cache[cache_key]
-            from_cache = True
-            logger.debug(
-                    "cache hit for generated scan kernel '%s'" % self.name_prefix)
-            (
-                self.first_level_scan_gen_info,
-                self.second_level_scan_gen_info,
-                self.final_update_gen_info) = result
-        except KeyError:
-            pass
+        if not cl._PYOPENCL_NO_CACHE:
+            try:
+                result = generic_scan_kernel_cache[cache_key]
+                from_cache = True
+                logger.debug(
+                    "cache hit for generated scan kernel '%s'", self.name_prefix)
+                (
+                    self.first_level_scan_gen_info,
+                    self.second_level_scan_gen_info,
+                    self.final_update_gen_info) = result
+            except KeyError:
+                pass
 
         if not from_cache:
             logger.debug(
-                    "cache miss for generated scan kernel '%s'" % self.name_prefix)
+                    "cache miss for generated scan kernel '%s'", self.name_prefix)
             self._finish_setup_impl()
 
             result = (self.first_level_scan_gen_info,
                       self.second_level_scan_gen_info,
                       self.final_update_gen_info)
 
-            generic_scan_kernel_cache.store_if_not_present(cache_key, result)
+            if not cl._PYOPENCL_NO_CACHE:
+                generic_scan_kernel_cache.store_if_not_present(cache_key, result)
 
         # Build the kernels.
         self.first_level_scan_info = self.first_level_scan_gen_info.build(
@@ -1210,7 +1225,7 @@ class GenericScanKernel(_GenericScanKernelBase):
                 self.context, self.options)
         del self.final_update_gen_info
 
-    def _finish_setup_impl(self):
+    def _finish_setup_impl(self) -> None:
         # {{{ find usable workgroup/k-group size, build first-level scan
 
         trip_count = 0
@@ -1223,8 +1238,8 @@ class GenericScanKernel(_GenericScanKernelBase):
             # not sure where these go, but roughly this much seems unavailable.
             avail_local_mem -= 0x400
 
-        is_cpu = self.devices[0].type & cl.device_type.CPU
-        is_gpu = self.devices[0].type & cl.device_type.GPU
+        is_cpu = bool(self.devices[0].type & cl.device_type.CPU)
+        is_gpu = bool(self.devices[0].type & cl.device_type.GPU)
 
         if is_cpu:
             # (about the widest vector a CPU can support, also taking
@@ -1250,7 +1265,7 @@ class GenericScanKernel(_GenericScanKernelBase):
         # k_group_size should be a power of two because of in-kernel
         # division by that number.
 
-        solutions = []
+        solutions: list[tuple[int, int, int]] = []
         for k_exp in range(0, 9):
             for wg_size in range(wg_size_multiples, max_scan_wg_size+1,
                     wg_size_multiples):
@@ -1262,7 +1277,6 @@ class GenericScanKernel(_GenericScanKernelBase):
                     solutions.append((wg_size*k_group_size, k_group_size, wg_size))
 
         if is_gpu:
-            from pytools import any
             for wg_size_floor in [256, 192, 128]:
                 have_sol_above_floor = any(wg_size >= wg_size_floor
                         for _, _, wg_size in solutions)
@@ -1316,10 +1330,12 @@ class GenericScanKernel(_GenericScanKernelBase):
         # {{{ build second-level scan
 
         from pyopencl.tools import VectorArg
-        second_level_arguments = self.parsed_args + [
-                VectorArg(self.dtype, "interval_sums")]
+        second_level_arguments = [
+            *self.parsed_args,
+            VectorArg(self.dtype, "interval_sums"),
+            ]
 
-        second_level_build_kwargs = {}
+        second_level_build_kwargs: dict[str, str | None] = {}
         if self.is_segmented:
             second_level_arguments.append(
                     VectorArg(self.index_dtype,
@@ -1364,9 +1380,10 @@ class GenericScanKernel(_GenericScanKernelBase):
             use_lookbehind_update=self.use_lookbehind_update,
             **self.code_variables))
 
-        update_scalar_arg_dtypes = (
-                get_arg_list_scalar_arg_dtypes(self.parsed_args)
-                + [self.index_dtype, self.index_dtype, None, None])
+        update_scalar_arg_dtypes = [
+                *get_arg_list_scalar_arg_dtypes(self.parsed_args),
+                self.index_dtype, self.index_dtype, None, None]
+
         if self.is_segmented:
             # g_first_segment_start_in_interval
             update_scalar_arg_dtypes.append(None)
@@ -1383,13 +1400,15 @@ class GenericScanKernel(_GenericScanKernelBase):
 
     # {{{ scan kernel build/properties
 
-    def get_local_mem_use(self, k_group_size, wg_size, use_bank_conflict_avoidance):
+    def get_local_mem_use(
+            self, k_group_size: int, wg_size: int,
+            use_bank_conflict_avoidance: bool) -> int:
         arg_dtypes = {}
         for arg in self.parsed_args:
             arg_dtypes[arg.name] = arg.dtype
 
-        fetch_expr_offsets = {}
-        for name, arg_name, ife_offset in self.input_fetch_exprs:
+        fetch_expr_offsets: dict[str, set[int]] = {}
+        for _name, arg_name, ife_offset in self.input_fetch_exprs:
             fetch_expr_offsets.setdefault(arg_name, set()).add(ife_offset)
 
         itemsize = self.dtype.itemsize
@@ -1411,17 +1430,24 @@ class GenericScanKernel(_GenericScanKernelBase):
                     for arg_name, ife_offsets in list(fetch_expr_offsets.items())
                     if -1 in ife_offsets or len(ife_offsets) > 1))
 
-    def generate_scan_kernel(self, max_wg_size, arguments, input_expr,
-            is_segment_start_expr, input_fetch_exprs, is_first_level,
-            store_segment_start_flags, k_group_size,
-            use_bank_conflict_avoidance):
+    def generate_scan_kernel(
+            self,
+            max_wg_size: int,
+            arguments: Sequence[DtypedArgument],
+            input_expr: str,
+            is_segment_start_expr: str | None,
+            input_fetch_exprs: Sequence[tuple[str, str, int]],
+            is_first_level: bool,
+            store_segment_start_flags: bool,
+            k_group_size: int,
+            use_bank_conflict_avoidance: bool) -> _GeneratedScanKernelInfo:
         scalar_arg_dtypes = get_arg_list_scalar_arg_dtypes(arguments)
 
         # Empirically found on Nv hardware: no need to be bigger than this size
         wg_size = _round_down_to_power_of_2(
                 min(max_wg_size, 256))
 
-        kernel_name = self.code_variables["name_prefix"]
+        kernel_name = cast("str", self.code_variables["name_prefix"])
         if is_first_level:
             kernel_name += "_lev1"
         else:
@@ -1460,7 +1486,25 @@ class GenericScanKernel(_GenericScanKernelBase):
 
     # }}}
 
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: Any, **kwargs: Any) -> cl.Event:
+        """
+        |std-enqueue-blurb|
+
+        .. note::
+
+            The returned :class:`pyopencl.Event` corresponds only to part of the
+            execution of the scan. It is not suitable for profiling.
+
+        :arg queue: queue on which to execute the scan. If not given, the
+            queue of the first :class:`pyopencl.array.Array` in *args* is used
+        :arg allocator: an allocator for the temporary arrays and results. If
+            not given, the allocator of the first :class:`pyopencl.array.Array`
+            in *args* is used.
+        :arg size: specify the length of the scan to be carried out. If not
+            given, this length is inferred from the first argument
+        :arg wait_for: a :class:`list` of events to wait for.
+        """
+
         # {{{ argument processing
 
         allocator = kwargs.get("allocator")
@@ -1474,8 +1518,8 @@ class GenericScanKernel(_GenericScanKernelBase):
             wait_for = list(wait_for)
 
         if len(args) != len(self.parsed_args):
-            raise TypeError("expected %d arguments, got %d" %
-                    (len(self.parsed_args), len(args)))
+            raise TypeError(
+                f"expected {len(self.parsed_args)} arguments, got {len(args)}")
 
         first_array = args[self.first_array_idx]
         allocator = allocator or first_array.allocator
@@ -1489,7 +1533,7 @@ class GenericScanKernel(_GenericScanKernelBase):
             return cl.enqueue_marker(queue, wait_for=wait_for)
 
         data_args = []
-        for arg_descr, arg_val in zip(self.parsed_args, args):
+        for arg_descr, arg_val in zip(self.parsed_args, args, strict=True):
             from pyopencl.tools import VectorArg
             if isinstance(arg_descr, VectorArg):
                 data_args.append(arg_val.base_data)
@@ -1514,29 +1558,30 @@ class GenericScanKernel(_GenericScanKernelBase):
 
         # {{{ allocate some buffers
 
-        interval_results = cl.array.empty(queue,
+        interval_results = cl_array.empty(queue,
                 num_intervals, dtype=self.dtype,
                 allocator=allocator)
 
-        partial_scan_buffer = cl.array.empty(
+        partial_scan_buffer = cl_array.empty(
                 queue, n, dtype=self.dtype,
                 allocator=allocator)
 
         if self.store_segment_start_flags:
-            segment_start_flags = cl.array.empty(
-                    queue, n, dtype=np.bool,
+            segment_start_flags = cl_array.empty(
+                    queue, n, dtype=np.bool_,
                     allocator=allocator)
 
         # }}}
 
         # {{{ first level scan of interval (one interval per block)
 
-        scan1_args = data_args + [
+        scan1_args = [
+                *data_args,
                 partial_scan_buffer.data, n, interval_size, interval_results.data,
                 ]
 
         if self.is_segmented:
-            first_segment_start_in_interval = cl.array.empty(queue,
+            first_segment_start_in_interval = cl_array.empty(queue,
                     num_intervals, dtype=self.index_dtype,
                     allocator=allocator)
             scan1_args.append(first_segment_start_in_interval.data)
@@ -1555,12 +1600,15 @@ class GenericScanKernel(_GenericScanKernelBase):
         # can scan at most one interval
         assert interval_size >= num_intervals
 
-        scan2_args = data_args + [
+        scan2_args = [
+                *data_args,
                 interval_results.data,  # interval_sums
                 ]
+
         if self.is_segmented:
             scan2_args.append(first_segment_start_in_interval.data)
-        scan2_args = scan2_args + [
+        scan2_args = [
+                *scan2_args,
                 interval_results.data,  # partial_scan_buffer
                 num_intervals, interval_size]
 
@@ -1572,7 +1620,8 @@ class GenericScanKernel(_GenericScanKernelBase):
 
         # {{{ update intervals with result of interval scan
 
-        upd_args = data_args + [
+        upd_args = [
+                *data_args,
                 n, interval_size, interval_results.data, partial_scan_buffer.data]
         if self.is_segmented:
             upd_args.append(first_segment_start_in_interval.data)
@@ -1653,8 +1702,17 @@ void ${name_prefix}_debug_scan(
 """
 
 
-class GenericDebugScanKernel(_GenericScanKernelBase):
-    def finish_setup(self):
+class GenericDebugScanKernel(GenericScanKernelBase):
+    """
+    Performs the same function and has the same interface as
+    :class:`GenericScanKernel`, but uses a dead-simple, sequential scan.  Works
+    best on CPU platforms, and helps isolate bugs in scans by removing the
+    potential for issues originating in parallel execution.
+
+    .. automethod:: __call__
+    """
+
+    def finish_setup(self) -> None:
         scan_tpl = _make_template(DEBUG_SCAN_TEMPLATE)
         scan_src = str(scan_tpl.render(
             output_statement=self.output_statement,
@@ -1668,15 +1726,17 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
             **self.code_variables))
 
         scan_prg = cl.Program(self.context, scan_src).build(self.options)
-        self.kernel = getattr(
-                scan_prg, self.name_prefix+"_debug_scan")
-        scalar_arg_dtypes = (
-                [None]
-                + get_arg_list_scalar_arg_dtypes(self.parsed_args)
-                + [self.index_dtype])
+        self.kernel = getattr(scan_prg, f"{self.name_prefix}_debug_scan")
+        scalar_arg_dtypes = [
+                None,
+                *get_arg_list_scalar_arg_dtypes(self.parsed_args),
+                self.index_dtype,
+                ]
         self.kernel.set_scalar_arg_dtypes(scalar_arg_dtypes)
 
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: Any, **kwargs: Any) -> cl.Event:
+        """See :meth:`GenericScanKernel.__call__`."""
+
         # {{{ argument processing
 
         allocator = kwargs.get("allocator")
@@ -1691,8 +1751,8 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
             wait_for = list(wait_for)
 
         if len(args) != len(self.parsed_args):
-            raise TypeError("expected %d arguments, got %d" %
-                    (len(self.parsed_args), len(args)))
+            raise TypeError(
+                f"expected {len(self.parsed_args)} arguments, got {len(args)}")
 
         first_array = args[self.first_array_idx]
         allocator = allocator or first_array.allocator
@@ -1701,13 +1761,13 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
         if n is None:
             n, = first_array.shape
 
-        scan_tmp = cl.array.empty(queue,
+        scan_tmp = cl_array.empty(queue,
                 n, dtype=self.dtype,
                 allocator=allocator)
 
         data_args = [scan_tmp.data]
         from pyopencl.tools import VectorArg
-        for arg_descr, arg_val in zip(self.parsed_args, args):
+        for arg_descr, arg_val in zip(self.parsed_args, args, strict=True):
             if isinstance(arg_descr, VectorArg):
                 data_args.append(arg_val.base_data)
                 if arg_descr.with_offset:
@@ -1718,8 +1778,7 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
 
         # }}}
 
-        return self.kernel(queue, (1,), (1,),
-                *(data_args + [n]), wait_for=wait_for)
+        return self.kernel(queue, (1,), (1,), *([*data_args, n]), wait_for=wait_for)
 
 # }}}
 
@@ -1729,11 +1788,11 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
 class _LegacyScanKernelBase(GenericScanKernel):
     def __init__(self, ctx, dtype,
             scan_expr, neutral=None,
-            name_prefix="scan", options=[], preamble="", devices=None):
+            name_prefix="scan", options=None, preamble="", devices=None):
         scan_ctype = dtype_to_ctype(dtype)
         GenericScanKernel.__init__(self,
                 ctx, dtype,
-                arguments="__global %s *input_ary, __global %s *output_ary" % (
+                arguments="__global {} *input_ary, __global {} *output_ary".format(
                     scan_ctype, scan_ctype),
                 input_expr="input_ary[i]",
                 scan_expr=scan_expr,
@@ -1752,8 +1811,8 @@ class _LegacyScanKernelBase(GenericScanKernel):
         if output_ary is None:
             output_ary = input_ary
 
-        if isinstance(output_ary, (str, six.text_type)) and output_ary == "new":
-            output_ary = cl.array.empty_like(input_ary, allocator=allocator)
+        if isinstance(output_ary, (str, str)) and output_ary == "new":
+            output_ary = cl_array.empty_like(input_ary, allocator=allocator)
 
         if input_ary.shape != output_ary.shape:
             raise ValueError("input and output must have the same shape")
@@ -1786,12 +1845,23 @@ class ExclusiveScanKernel(_LegacyScanKernelBase):
 # {{{ template
 
 class ScanTemplate(KernelTemplateBase):
-    def __init__(self,
-            arguments, input_expr, scan_expr, neutral, output_statement,
-            is_segment_start_expr=None, input_fetch_exprs=[],
-            name_prefix="scan", preamble="", template_processor=None):
+    def __init__(
+            self,
+            arguments: str | list[DtypedArgument],
+            input_expr: str,
+            scan_expr: str,
+            neutral: str | None,
+            output_statement: str,
+            is_segment_start_expr: str | None = None,
+            input_fetch_exprs: list[tuple[str, str, int]] | None = None,
+            name_prefix: str = "scan",
+            preamble: str = "",
+            template_processor: Any = None) -> None:
+        super().__init__(template_processor=template_processor)
+
+        if input_fetch_exprs is None:
+            input_fetch_exprs = []
 
-        KernelTemplateBase.__init__(self, template_processor=template_processor)
         self.arguments = arguments
         self.input_expr = input_expr
         self.scan_expr = scan_expr
@@ -1804,7 +1874,7 @@ class ScanTemplate(KernelTemplateBase):
 
     def build_inner(self, context, type_aliases=(), var_values=(),
             more_preamble="", more_arguments=(), declare_types=(),
-            options=(), devices=None, scan_cls=GenericScanKernel):
+            options=None, devices=None, scan_cls=GenericScanKernel):
         renderer = self.get_renderer(type_aliases, var_values, context, options)
 
         arg_list = renderer.render_argument_list(self.arguments, more_arguments)
@@ -1819,7 +1889,7 @@ class ScanTemplate(KernelTemplateBase):
             is_segment_start_expr=renderer(self.is_segment_start_expr),
             input_fetch_exprs=self.input_fetch_exprs,
             index_dtype=renderer.type_aliases.get("index_t", np.int32),
-            name_prefix=renderer(self.name_prefix), options=list(options),
+            name_prefix=renderer(self.name_prefix), options=options,
             preamble=(
                 type_decl_preamble
                 + "\n"
